#!/usr/bin/env python3
"""
Utility functions for TCGA REST API.
"""
import sys,os,re,json,time,logging

from ..util import rest
#
##############################################################################
def ListProjects(base_url, skip, nmax, fout):
  n_out=0; tags=None;
  from_next=skip; size=100;
  while True:
    url_next = (base_url+'/projects?from={0}&size={1}'.format(from_next, size))
    rval = rest.GetURL(url_next, parse_json=True)
    projects = rval["data"]["hits"] if "data" in rval and "hits" in rval["data"] else []
    for project in projects:
      logging.debug(json.dumps(project, indent=2))
      if not tags:
        tags = list(project.keys())
        fout.write("\t".join(tags)+"\n")
      vals = [(str(project[tag]) if tag in project else "") for tag in tags]
      fout.write("\t".join(vals)+"\n")
      n_out+=1
      if n_out>=nmax: break
    if n_out>=nmax: break
    total = rval["data"]["pagination"]["total"] if "data" in rval and "pagination" in rval["data"] and "total" in rval["data"]["pagination"] else None
    count = rval["data"]["pagination"]["count"] if "data" in rval and "pagination" in rval["data"] and "count" in rval["data"]["pagination"] else None
    if not count or count<size: break
    from_next += count
  logging.info("n_out: %d / %d"%(n_out, total))

##############################################################################
def ListCases(base_url, skip, nmax, fout):
  n_out=0; tags=None;
  from_next=skip; size=100;
  while True:
    url_next = (base_url+'/cases?from={0}&size={1}'.format(from_next, size))
    rval = rest.GetURL(url_next, parse_json=True)
    cases = rval["data"]["hits"] if "data" in rval and "hits" in rval["data"] else []
    for case in cases:
      logging.debug(json.dumps(case, indent=2))
      if not tags:
        tags = list(case.keys())
        fout.write("\t".join(tags)+"\n")
      vals = [(str(case[tag]) if tag in case else "") for tag in tags]
      fout.write("\t".join(vals)+"\n")
      n_out+=1
      if n_out>=nmax: break
    if n_out>=nmax: break
    total = rval["data"]["pagination"]["total"] if "data" in rval and "pagination" in rval["data"] and "total" in rval["data"]["pagination"] else None
    count = rval["data"]["pagination"]["count"] if "data" in rval and "pagination" in rval["data"] and "count" in rval["data"]["pagination"] else None
    if not count or count<size: break
    from_next += count
  logging.info("n_out: %d / %d"%(n_out, total))

##############################################################################
def ListFiles(base_url, skip, nmax, fout):
  n_out=0; tags=None;
  from_next=skip; size=100;
  while True:
    url_next = (base_url+'/files?from={0}&size={1}'.format(from_next, size))
    rval = rest.GetURL(url_next, parse_json=True)
    files = rval["data"]["hits"] if "data" in rval and "hits" in rval["data"] else []
    for file_this in files:
      logging.debug(json.dumps(file_this, indent=2))
      if not tags:
        tags = list(file_this.keys())
        fout.write("\t".join(tags)+"\n")
      vals = [(str(file_this[tag]) if tag in file_this else "") for tag in tags]
      fout.write("\t".join(vals)+"\n")
      n_out+=1
      if n_out>=nmax: break
    if n_out>=nmax: break
    total = rval["data"]["pagination"]["total"] if "data" in rval and "pagination" in rval["data"] and "total" in rval["data"]["pagination"] else None
    count = rval["data"]["pagination"]["count"] if "data" in rval and "pagination" in rval["data"] and "count" in rval["data"]["pagination"] else None
    if not count or count<size: break
    from_next += count
  logging.info("n_out: %d / %d"%(n_out, total))

##############################################################################
def ListAnnotations(base_url, skip, nmax, fout):
  n_out=0; tags=None;
  from_next=skip; size=100;
  while True:
    url_next = (base_url+'/annotations?from={0}&size={1}'.format(from_next, size))
    rval = rest.GetURL(url_next, parse_json=True)
    annos = rval["data"]["hits"] if "data" in rval and "hits" in rval["data"] else []
    for anno in annos:
      logging.debug(json.dumps(anno, indent=2))
      if not tags:
        tags = list(anno.keys())
        fout.write("\t".join(tags)+"\n")
      vals = [(str(anno[tag]) if tag in anno else "") for tag in tags]
      fout.write("\t".join(vals)+"\n")
      n_out+=1
      if n_out>=nmax: break
    if n_out>=nmax: break
    total = rval["data"]["pagination"]["total"] if "data" in rval and "pagination" in rval["data"] and "total" in rval["data"]["pagination"] else None
    count = rval["data"]["pagination"]["count"] if "data" in rval and "pagination" in rval["data"] and "count" in rval["data"]["pagination"] else None
    if not count or count<size: break
    from_next += count
  logging.info("n_out: %d / %d"%(n_out, total))

##############################################################################



#!/usr/bin/env python3
"""
OpenFDA Adverse Events Reports REST API utility functions.

 Reports include multiple reactions with multiple drugs, i.e. many-to-many
 drug-reaction associations.  It is inherently unknown which drug or drug
 combination is explanatory.

 My API key: ==SEE $HOME/.fda.yaml==
 My email: ==SEE $HOME/.fda.yaml==
 
 Query params:
   search	(e.g. search=patient.drug.drugindication:"multiple+myeloma")
   count	(e.g. count=patient.drug.drugindication)
   limit	(e.g. limit=25)
   skip	(e.g. skip=100)
 
 https://api.fda.gov/drug/event.json?search=patient.drug.openfda.pharm_class_epc:"nonsteroidal+anti-inflammatory+drug"&count=patient.reaction.reactionmeddrapt.exact
 
 patient.reaction.reactionmeddrapt  - MedDRA term(s) for the reaction(s).
 
 patient.reaction.reactionmeddraversionpt - The MedDRA version that patient.reaction.reactionmeddrapt uses.
 
 patient.reaction.reactionoutcome - Outcome of the reaction/event at the time of last observation
	1 = recovered/resolved
	2 = recovering/resolving
	3 = not recovered/not resolved
	4 = recovered/resolved with sequelae
	5 = fatal
	6 = unknown
"""
import sys,os,re,time,json,logging
import pandas as pd
#
from ...util import rest
#
REST_RETRY_NMAX=10
REST_RETRY_WAIT=5

DRUG_ID_FIELDS = ['unii','rxcui','nui','spl_id','product_ndc','package_ndc']
DRUG_NAME_FIELDS = ['generic_name','brand_name','substance_name']

API_HOST='api.fda.gov'
API_BASE_PATH='/drug/event.json'
API_BASE_URL='https://'+API_HOST+API_BASE_PATH

##############################################################################
def GetCounts(tfrom, tto, base_url=API_BASE_URL):
  txt=''
  try:
    rval=rest.Utils.GetURL(base_url+f"?search=receivedate:[{tfrom}+TO+{tto}]&count=receivedate")
    txt+=(f"{rval}\n")
  except Exception as e:
    logging.error(str(e))
  return txt

#############################################################################
def Info(base_url=API_BASE_URL):
  url=(base_url+'?search=(serious:1)&limit=1')
  try:
    rval=rest.Utils.GetURL(url, parse_json=True)
  except Exception as e:
    logging.error(str(e))
    return None
  meta = rval['meta']
  return meta

#############################################################################
def GetFields(base_url=API_BASE_URL):
  '''Show fields where value is either (1) string, or (2) list of strings.
Sample first n records.  However no guarantee sampling contains all fields.
'''
  n=100
  url = (base_url+f"?search=(serious:1)&limit={n}")
  try:
    rval=rest.Utils.GetURL(url,parse_json=True)
  except Exception as e:
    logging.error(str(e))
    return None
  logging.info(f"Fields from sampled records, N = {n}")
  fields=set()
  for result in rval['results']:
    fields |= GetFieldsIn(result,'')
  return sorted(list(fields))

#############################################################################
def GetFieldsIn(obj, path):
  fields=set()
  if type(obj) is str: return
  elif type(obj) is list:
    if len(obj)==0 or type(obj[0]) is str: return
    for obj2 in obj:
      fields|=GetFieldsIn(obj2,path)
  elif type(obj) is dict:
    for field in obj.keys():
      path_this = (f"{path}{'.' if path else ''}{field}")
      if type(obj[field]) is str:
        fields.add(path_this)
      elif type(obj[field]) is list:
        if len(obj[field])>0:
          if type(obj[field][0]) is str:
            fields.add(path_this)
          else:
            for obj2 in obj[field]:
              fields|=(GetFieldsIn(obj2, path_this))
      else:
        fields|=GetFieldsIn(obj[field], path_this)
  return fields

#############################################################################
def DrugName(drug):
  name=''
  if 'openfda' in drug :
    for name_field in DRUG_NAME_FIELDS:
      if name_field in drug['openfda'] :
        if len(drug['openfda'][name_field])>0:
          name=drug['openfda'][name_field][0]
          break
    if not name:
      for id_field in DRUG_ID_FIELDS:
        if id_field in drug['openfda']:
          if len(drug['openfda'][id_field])>0:
            name = (f"{id_field}={drug['openfda'][id_field][0]}")
            break
  else:
    if 'medicinalproduct' in drug :
      name=drug['medicinalproduct']
  return name

#############################################################################
def DrugID(drug, id_field):
  if 'openfda' in drug :
    if id_field in drug['openfda'] :
      if len(drug['openfda'][id_field])>0:
        return drug['openfda'][id_field][0]
  return ''

#############################################################################
def Search(drug_cl, drug_ind, drug_unii, drug_ndc, drug_spl, tfrom, tto, serious, fatal, rawquery, nmax, api_key, base_url=API_BASE_URL, fout=None):
  '''The API seems to disallow limit exceeding 100.  So we need to iterate using skip.'''
  rval=None; qrys=[]; df=None; tags=[]; n_out=0;
  tags_result=[]; tags_patient=[]; tags_drug=[]; tags_reaction=[];
  url = (base_url+f"?api_key={api_key}&search=")
  if drug_cl:	qrys.append(f"(patient.drug.openfda.pharm_class_epc:{drug_cl.replace(' ','+')})")
  if drug_ind:	qrys.append(f"(patient.drug.drugindication:{drug_ind.replace(' ','+')})")
  if drug_unii:	qrys.append(f"(patient.drug.openfda.unii:{drug_unii})")
  if drug_ndc:	qrys.append(f"(patient.drug.openfda.ndc:{drug_ndc})")
  if drug_spl:	qrys.append(f"(patient.drug.openfda.spl:{drug_spl})")
  if fatal:	qrys.append(f"(seriousnessdeath:1)")
  if serious:	qrys.append(f"(serious:1)")
  if tfrom:	qrys.append(f"(receivedate:[{tfrom}+TO+{tto}])")
  if rawquery:	qrys.append(f"({rawquery.replace(' ','+')})")
  if len(qrys)==0:
    logging.error('No query specified.')
    return rval
  url+=('+AND+'.join(qrys))
  drugnames_all = set(); uniis_all = set(); rxns_all = set()
  ser_counts =  {s:0 for s in ('F','H','S','')}
  fromtime = time.strptime('19000101','%Y%m%d')
  totime = time.localtime()
  ndone=0; nchunk=100; n_report=0;
  while nmax==0 or ndone<nmax:
    if nmax>ndone: nchunk = min(nchunk, nmax-ndone)
    url_this = url+(f"&limit={nchunk}")+(f"&skip={ndone}" if ndone>0 else '')
    rval = rest.Utils.GetURL(url_this, parse_json=True)
    if not rval: break
    results = rval['results']
    logging.debug(f"N_results_this: {len(results)}")
    for result in results:
      if not 'drug' in result['patient']:
        logging.warning(f"Report [ID={result['safetyreportid']}] missing drug[s].")
        continue
      n_report+=1
      if not tags_result:
        for tag in result.keys():
          if type(result[tag]) not in (list, dict):
            tags_result.append(tag) #Only simple metadata.
      df_result = pd.DataFrame({tags_result[j]:[result[tags_result[j]]] if tags_result[j] in result else [None] for j in range(len(tags_result))})
      ser = ('F' if 'seriousnessdeath' in result else ('H' if 'seriousnesshospitalization' in result else ('S' if 'serious' in result else '')))
      ser_counts[ser]+=1
      fromtime = min(fromtime, time.strptime(result['receiptdate'],'%Y%m%d'))
      totime = max(totime, time.strptime(result['receiptdate'],'%Y%m%d'))
      patient = result['patient']
      logging.debug(json.dumps(patient, indent=2))
      if not tags_patient:
        for tag in patient.keys():
          if type(patient[tag]) not in (list, dict):
            tags_patient.append(tag) #Only simple metadata.
      df_patient = pd.DataFrame({tags_patient[j]:[patient[tags_patient[j]]] if tags_patient[j] in patient else [None] for j in range(len(tags_patient))})
      reactions = patient['reaction']
      drugs = patient['drug']
      drugnames = set(); uniis = set(); rxns = set()
      for drug in drugs:
        logging.debug(json.dumps(drug, indent=2))
        if not tags_drug:
          for tag in drug.keys():
            if type(drug[tag]) not in (list, dict):
              tags_drug.append(tag) #Only simple metadata.
        df_drug = pd.DataFrame({tags_drug[j]:[drug[tags_drug[j]]] if tags_drug[j] in drug else [None] for j in range(len(tags_drug))})
        drugnames.add(DrugName(drug))
        uniis.add(DrugID(drug, 'unii'))
        for reaction in reactions:
          logging.debug(json.dumps(reaction, indent=2))
          if not tags_reaction:
            for tag in reaction.keys():
              if type(reaction[tag]) not in (list, dict):
                tags_reaction.append(tag) #Only simple metadata.
          df_reaction = pd.DataFrame({tags_reaction[j]:[reaction[tags_reaction[j]]] if tags_reaction[j] in reaction else [None] for j in range(len(tags_reaction))})

#          df_this = pd.DataFrame(
#		{"ReportID":[result['safetyreportid']],
#		"ReceiptDate":[result['receiptdate']],
#		"Drugname":[DrugName(drug)],
#		"UNII":[DrugID(drug, 'unii')],
#		"ProductNDC":[DrugID(drug, 'product_ndc')],
#		"Seriousness":[ser],
#		"Event":[reaction['reactionmeddrapt']]})

          df_drug_extra = pd.DataFrame({
		"Drugname":[DrugName(drug)],
		"UNII":[DrugID(drug, 'unii')],
		"ProductNDC":[DrugID(drug, 'product_ndc')]})

          df_this = pd.concat([df_result, df_patient, df_drug_extra, df_drug, df_reaction], axis=1)
          #df_this = df_this.dropna(how="all")

          rxns.add(reaction['reactionmeddrapt'])
          if fout is None: df = pd.concat([df, df_this])
          else: df_this.to_csv(fout, "\t", index=False, header=bool(n_out==0))
          n_out += df_this.shape[0]
      logging.debug(f"{n_report}. Report: {result['safetyreportid']} [{result['receiptdate']}] seriousness: {ser}; reactions: {(', '.join(list(rxns)))}; drugs: {(', '.join(list(drugnames)))}")
      uniis_all |= uniis
      drugnames_all |= drugnames
      rxns_all |= rxns
    ndone+=nchunk
    if nmax>0 and ndone>=nmax: break

  logging.info(f"n_out: {n_out}")
  logging.info(f"N_report: {n_report}; drugs: {len(uniis_all)}; reactions: {len(rxns_all)}")
  logging.info(f"Seriousness: {str(ser_counts)}; total: {sum(ser_counts.values())}")
  logging.info(f"Daterange: ({time.strftime('%Y%m%d',fromtime)}-{time.strftime('%Y%m%d',totime)})")
  if fout is None:
    return df



#!/usr/bin/env python3
"""
https://mor.nlm.nih.gov/download/rxnav/RxNormAPIs.html
https://www.nlm.nih.gov/research/umls/rxnorm/docs/
"""
###
import sys,os,re,json,logging,urllib.parse,tqdm
import pandas as pd

from ..util import rest

#
API_HOST='rxnav.nlm.nih.gov'
API_BASE_PATH='/REST'
BASE_URL='https://'+API_HOST+API_BASE_PATH
#
NDFRT_TYPES=('DISEASE','INGREDIENT','MOA','PE','PK') ## NDFRT drug class types
#
#############################################################################
def List_IDTypes(base_url=BASE_URL, fout=None):
  rval = rest.GetURL(base_url+'/idtypes.json', parse_json=True)
  logging.debug(json.dumps(rval, indent=2))
  df = pd.DataFrame({"idTypeList":rval['idTypeList']['idName']})
  logging.info(f"n_out: {df.shape[0]}")
  if fout is not None: df.to_csv(fout, "\t", index=False)
  else: return df

#############################################################################
def List_SourceTypes(base_url=BASE_URL, fout=None):
  rval = rest.GetURL(base_url+'/sourcetypes.json', parse_json=True)
  df = pd.DataFrame({"sourceType":rval['sourceTypeList']['sourceName']})
  logging.info(f"n_out: {df.shape[0]}")
  if fout is not None: df.to_csv(fout, "\t", index=False)
  else: return df

#############################################################################
def List_RelationTypes(base_url=BASE_URL, fout=None):
  rval = rest.GetURL(base_url+'/relatypes.json', parse_json=True)
  df = pd.DataFrame({"relationType":rval['relationTypeList']['relationType']})
  logging.info(f"n_out: {df.shape[0]}")
  if fout is not None: df.to_csv(fout, "\t", index=False)
  else: return df

#############################################################################
def List_TermTypes(base_url=BASE_URL, fout=None):
  rval = rest.GetURL(base_url+'/termtypes.json', parse_json=True)
  df = pd.DataFrame({"termType":rval['termTypeList']['termType']})
  logging.info(f"n_out: {df.shape[0]}")
  if fout is not None: df.to_csv(fout, "\t", index=False)
  else: return df

#############################################################################
def List_PropNames(base_url=BASE_URL, fout=None):
  rval = rest.GetURL(base_url+'/propnames.json', parse_json=True)
  df = pd.DataFrame({"propName":rval['propNameList']['propName']})
  logging.info(f"n_out: {df.shape[0]}")
  if fout is not None: df.to_csv(fout, "\t", index=False)
  else: return df

#############################################################################
def List_PropCategories(base_url=BASE_URL, fout=None):
  rval = rest.GetURL(base_url+'/propCategories.json', parse_json=True)
  df = pd.DataFrame({"propCategory":rval['propCategoryList']['propCategory']})
  logging.info(f"n_out: {df.shape[0]}")
  if fout is not None: df.to_csv(fout, "\t", index=False)
  else: return df

#############################################################################
def List_ClassTypes(base_url=BASE_URL, fout=None):
  rval = rest.GetURL(base_url+'/rxclass/classTypes.json', parse_json=True)
  df = pd.DataFrame({"classType":rval['classTypeList']['classTypeName']})
  logging.info(f"n_out: {df.shape[0]}")
  if fout is not None: df.to_csv(fout, "\t", index=False)
  else: return df

#############################################################################
def List_Classes(class_types, base_url=BASE_URL, fout=None):
  n_out=0; tags=None; df=pd.DataFrame(); tq=None;
  url = (f'{base_url}/rxclass/allClasses.json')
  if class_types: url+=("?classTypes="+urllib.parse.quote(' '.join(class_types)))
  rval = rest.GetURL(url, parse_json=True)
  logging.debug(json.dumps(rval, indent=4))
  clss = rval["rxclassMinConceptList"]["rxclassMinConcept"] if "rxclassMinConceptList" in rval and "rxclassMinConcept" in rval["rxclassMinConceptList"] else []
  for cls in clss:
    if not tq: tq = tqdm.tqdm(total=len(clss), unit="classes")
    tq.update()
    if not tags: tags = list(cls.keys())
    df_this = pd.DataFrame({tags[j]:[cls[tags[j]]] for j in range(len(tags))})
    if fout is None: df = pd.concat([df, df_this])
    else: df_this.to_csv(fout, "\t", index=False, header=bool(n_out==0))
    n_out += df_this.shape[0]
  logging.info(f"n_out: {n_out}")
  if fout is None: return df

#############################################################################
def Get_Name2RxCUI(names, base_url=BASE_URL, fout=None):
  n_out=0; df=pd.DataFrame();
  for name in names:
    rval = rest.GetURL(f'{base_url}/rxcui.json?name={urllib.parse.quote(name)}', parse_json=True)
    logging.debug(json.dumps(rval, indent=4))
    idGroup = rval["idGroup"] if "idGroup" in rval else None
    rxnormIds = idGroup["rxnormId"] if idGroup and "rxnormId" in idGroup else []
    for rxnormId in rxnormIds:
      df_this = pd.DataFrame({"name":[idGroup["name"]], "rxnormId":rxnormId})
      if fout is None: df = pd.concat([df, df_this])
      else: df_this.to_csv(fout, "\t", index=False, header=bool(n_out==0))
      n_out += df_this.shape[0]
  logging.info(f"n_out: {n_out}")
  if fout is None: return df

#############################################################################
def Get_Name(names, base_url=BASE_URL, fout=None):
  n_out=0; tags=None; df=pd.DataFrame();
  for name in names:
    rval = rest.GetURL(f'{base_url}/drugs.json?name={urllib.parse.quote(name)}', parse_json=True)
    logging.debug(json.dumps(rval, indent=4))
    conceptGroups = rval["drugGroup"]["conceptGroup"] if "drugGroup" in rval and "conceptGroup" in rval["drugGroup"] else []
    for cgroup in conceptGroups:
      cprops = cgroup["conceptProperties"] if "conceptProperties" in cgroup else []
      if not cprops: continue
      for cprop in cprops:
        if not tags: tags = list(cprop.keys())
        df_this = pd.DataFrame({tags[j]:[cprop[tags[j]]] for j in range(len(tags))})
        if fout is None: df = pd.concat([df, df_this])
        else: df_this.to_csv(fout, "\t", index=False, header=bool(n_out==0))
        n_out += df_this.shape[0]
  logging.info(f"n_out: {n_out}")
  if fout is None: return df

#############################################################################
def Get_ID2RxCUI(ids, idtype, base_url=BASE_URL, fout=None):
  """For mapping external ID, for supported ID types, to RxNorm ID."""
  n_out=0; df=pd.DataFrame();
  for id_this in ids:
    rval = rest.GetURL(f'{base_url}/rxcui.json?idtype={idtype}&id={id_this}', parse_json=True)
    logging.debug(json.dumps(rval, indent=4))
    for rxcui in rval['idGroup']['rxnormId']:
      df_this = pd.DataFrame({"idtype":idtype, "id":[id_this], "rxnormId":rxcui})
      if fout is None: df = pd.concat([df, df_this])
      else: df_this.to_csv(fout, "\t", index=False, header=bool(n_out==0))
      n_out += df_this.shape[0]
  logging.info(f"n_out: {n_out}")
  if fout is None: return df

#############################################################################
def Get_RxCUI_Status(ids, base_url=BASE_URL, fout=None):
  n_out=0;
  tags = ["rxcui", "name", "tty"]
  fout.write('\t'.join(tags)+"\tstatus\tremappedDate\n")
  for rxcui in ids:
    rval = rest.GetURL(f'{base_url}/rxcui/{rxcui}/status.json', parse_json=True)
    logging.debug(json.dumps(rval, indent=4))
    if "rxcuiStatus" not in rval :
      logging.error(f"Bad record: {rval}")
      continue
    status = rval["rxcuiStatus"]["status"] if "status" in rval["rxcuiStatus"] else ""
    remappedDate = rval["rxcuiStatus"]["remappedDate"] if "remappedDate" in rval["rxcuiStatus"] else ""
    minConcepts = rval["rxcuiStatus"]["minConceptGroup"]["minConcept"] if "minConceptGroup" in rval["rxcuiStatus"] and "minConcept" in rval["rxcuiStatus"]["minConceptGroup"] else []
    for minConcept in minConcepts:
      vals = [minConcept[tag] if tag in minConcept else '' for tag in tags]
      fout.write('\t'.join(vals+[status, remappedDate])+"\n")
    n_out+=1
  logging.info(f"n_out: {n_out}")

#############################################################################
def Get_RxCUI_Properties(ids, base_url=BASE_URL, fout=None):
  n_out=0; tags=None;
  for rxcui in ids:
    rval = rest.GetURL(f'{base_url}/rxcui/{rxcui}/properties.json', parse_json=True)
    logging.debug(json.dumps(rval, indent=4))
    props = rval["properties"] if "properties" in rval else []
    if not tags:
      tags = props.keys()
      fout.write('\t'.join(tags)+'\n')
    vals = [props[tag] if tag in props else '' for tag in tags]
    fout.write('\t'.join(vals)+'\n')
    n_out+=1
  logging.info(f"n_out: {n_out}")

#############################################################################
def Get_RxCUI_AllProperties(ids, base_url=BASE_URL, fout=None):
  n_out=0; tags=None;
  for rxcui in ids:
    rval = rest.GetURL(f'{base_url}/rxcui/{rxcui}/allProperties.json?prop=all', parse_json=True)
    logging.debug(json.dumps(rval, indent=4))
    props = rval["propConceptGroup"]["propConcept"] if "propConceptGroup" in rval and "propConcept" in rval["propConceptGroup"] else {}
    for prop in props:
      if not tags:
        tags = list(prop.keys())
        fout.write('\t'.join(['RxCui']+tags)+'\n')
      vals = [prop[tag] if tag in prop else '' for tag in tags]
      fout.write('\t'.join([rxcui]+vals)+'\n')
      n_out+=1
  logging.info(f"n_out: {n_out}")

#############################################################################
def Get_RxCUI_AllRelated(ids, base_url=BASE_URL, fout=None):
  n_out=0; tags=None;
  for rxcui in ids:
    rval = rest.GetURL(f'{base_url}/rxcui/{rxcui}/allrelated.json', parse_json=True)
    logging.debug(json.dumps(rval, indent=4))
    conceptGroups = rval["allRelatedGroup"]["conceptGroup"] if "allRelatedGroup" in rval and "conceptGroup" in rval["allRelatedGroup"] else []
    for cgroup in conceptGroups:
      props = cgroup["conceptProperties"] if "conceptProperties" in cgroup else []
      for prop in props:
        if not tags:
          tags = list(prop.keys())
          fout.write('\t'.join(['RxCui']+tags)+'\n')
        vals = [prop[tag] if tag in prop else '' for tag in tags]
        fout.write("\t".join([rxcui]+vals)+"\n")
        n_out+=1
  logging.info(f"n_out: {n_out}")

#############################################################################
def Get_RxCUI_NDCs(ids, base_url=BASE_URL, fout=None):
  n_out=0;
  fout.write('rxcui\tndc\n')
  for rxcui in ids:
    rval = rest.GetURL(f'{base_url}/rxcui/{rxcui}/ndcs.json', parse_json=True)
    ndcs = rval['ndcGroup']['ndcList']['ndc'] if 'ndcGroup' in rval and 'ndcList' in rval['ndcGroup'] and rval['ndcGroup']['ndcList'] and 'ndc' in rval['ndcGroup']['ndcList'] else []
    for ndc in ndcs:
      fout.write('%s\t%s\n'%(rxcui, ndc))
      n_out+=1
  logging.info(f"n_out: {n_out}")

#############################################################################
def Get_RxCUI_Classes(ids, base_url=BASE_URL, fout=None):
  n_out=0;
  tags = ["classId", "classType", "className" ]
  for rxcui in ids:
    rval = rest.GetURL(f'{base_url}/rxcui/{rxcui}/classes.json', parse_json=True)
    logging.debug(json.dumps(rval, indent=4))

#############################################################################
def Get_Class_Members(class_id, rel_src, base_url=BASE_URL, fout=None):
  n_out=0;
  url = (f'{base_url}/rxclass/classMembers.json?classId={class_id}&relaSource={rel_src}')
  rval = rest.GetURL(url, parse_json=True)
  logging.debug(json.dumps(rval, indent=4))
  #???
  logging.info(f"n_out: {n_out}")

#############################################################################
def Get_SpellingSuggestions(name, base_url=BASE_URL, fout=None):
  rval = rest.GetURL(f'{base_url}/spellingsuggestions.json?name={urllib.parse.quote(name)}', parse_json=True)
  logging.debug(json.dumps(rval, indent=4))
  #???

#############################################################################
def RawQuery(rawquery, base_url=BASE_URL, fout=None):
  rval = rest.GetURL(base_url+rawquery, parse_json=True)
  logging.debug(json.dumps(rval, indent=4))
  #???




#!/usr/bin/env python3
"""
RDF utility functions.
"""
import sys,os,re,gzip,argparse,logging

import rdflib

from .. import rdf as util_rdf

#############################################################################
if __name__=="__main__":
  parser = argparse.ArgumentParser(description="RDF utility", epilog="")
  ops = [ "describe_rdf", "validate_rdf", "convert_rdf", "describe_owl", "validate_owl", ]
  FORMATS = ["text/turtle", "application/rdf+xml", "text/n3", ]
  parser.add_argument("op", choices=ops, help="OPERATION")
  parser.add_argument("--i", dest="ifile", help="input file (RDF or OWL)")
  parser.add_argument("--ifmt", choices=FORMATS, default="text/turtle", help="input RDF format")
  parser.add_argument("--ofmt", choices=FORMATS, default="text/turtle", help="output RDF format")
  parser.add_argument("--o", dest="ofile", help="output file")
  parser.add_argument("-v", "--verbose", action="count", default=0)
  args = parser.parse_args()

  logging.basicConfig(format="%(levelname)s:%(message)s", level=(logging.DEBUG if args.verbose>1 else logging.INFO))

  fin = open(args.ifile, "r") if args.ifile else sys.stdin
  fout = open(args.ofile, "w") if args.ofile else sys.stdout

  if args.op == "describe_rdf":
    util_rdf.DescribeRdf(fin, args.ifmt)

  elif args.op == "validate_rdf":
    util_rdf.ValidateRdf(fin, args.ifmt)

  elif args.op == "convert_rdf":
    util_rdf.ConvertRdf(fin, args.ifmt, args.ofmt, fout)

  elif args.op == "describe_owl":
    util_rdf.DescribeOwl(args.ifile)

  elif args.op == "validate_owl":
    util_rdf.ValidateOwl(args.ifile)

  else:
    parser.error(f"Invalid operation: {args.op}")



#!/usr/bin/env python3
"""
RDF utility functions.
https://rdflib.readthedocs.io/
https://owlready2.readthedocs.io/
"""
import sys,os,re,gzip,argparse,logging

import rdflib
import owlready2

#############################################################################
def LoadRdfFile(fin, ifmt):
  g = rdflib.Graph()
  try:
    g.parse(fin, format=ifmt)
    logging.info(f"RDF graph from {fin.name} ({ifmt}) contains {len(g)} triples.")
  except Exception as e:
    logging.error(e)
    return None
  return g

#############################################################################
def ValidateRdf(fin, ifmt):
  g = LoadRdfFile(fin, ifmt)
  if g is not None:
    logging.info(f"RDF file VALIDATED: {fin.name} ({ifmt})")
    return True
  else:
    logging.info(f"RDF file NOT VALIDATED: {fin.name} ({ifmt})")
    return False

#############################################################################
def DescribeRdf(fin, ifmt):
  g = LoadRdfFile(fin, ifmt)

#############################################################################
def ConvertRdf(fin, ifmt, ofmt, fout):
  g = LoadRdfFile(fin, ifmt)
  fout.write(g.serialize(format=ofmt).decode("utf8"))
  logging.info(f"RDF graph to {fout.name} ({ofmt}) containing {len(g)} triples.")

#############################################################################
def LoadOwlFile(ifile):
  try:
    onto = owlready2.get_ontology(f"file://{ifile}").load()
    logging.info(f"OWL ontology from {ifile} contains {len(list(onto.classes()))} classes.")
  except Exception as e:
    logging.error(e)
    return None
  return onto

#############################################################################
def DescribeOwl(ifile):
  onto = LoadOwlFile(ifile)

#############################################################################
def ValidateOwl(ifile):
  onto = LoadOwlFile(ifile)
  if onto is not None:
    logging.info(f"OWL file VALIDATED: {ifile}")
    return True
  else:
    logging.info(f"OWL file NOT VALIDATED: {ifile}")
    return False

#############################################################################



#!/usr/bin/env python3
##############################################################################
### MedlinePlus Connect utilities - access SNOMED and ICD codes
### https://medlineplus.gov/connect/technical.html
### https://medlineplus.gov/connect/service.html
##############################################################################
### https://apps.nlm.nih.gov/medlineplus/services/mpconnect_service.cfm
### Two required parameters:
### 1. Code System (one of): 
###	ICD-10-CM: mainSearchCriteria.v.cs=2.16.840.1.113883.6.90
###	ICD-9-CM: mainSearchCriteria.v.cs=2.16.840.1.113883.6.103
###	SNOMED_CT: mainSearchCriteria.v.cs=2.16.840.1.113883.6.96
###	NDC: mainSearchCriteria.v.cs=2.16.840.1.113883.6.69
###	RXNORM: mainSearchCriteria.v.cs=2.16.840.1.113883.6.88
###	LOINC:	mainSearchCriteria.v.cs=2.16.840.1.113883.6.1
### 2. Code: 
###	mainSearchCriteria.v.c=250.33
###
### Content format:
###	XML (default): knowledgeResponseType=text/xml
###	JSON: knowledgeResponseType=application/json
###	JSONP: knowledgeResponseType=application/javascript&callback=CallbackFunction
###	  where CallbackFunction is a name you give the call back function.
##############################################################################
import sys,os,re,argparse,time,logging
#
from ... import medline
#
##############################################################################
if __name__=='__main__':
  ops = ["get_code"]
  epilog = f"Supported codesystems: {sorted(medline.connect.Utils.CODESYSTEMS.keys())}"
  parser = argparse.ArgumentParser(description='MedlinePlus REST API client', epilog=epilog)
  parser.add_argument("op", choices=ops, help='OPERATION (select one)')
  parser.add_argument("--i", dest="ifile", help="input code ID file (one per line)")
  parser.add_argument("--o", dest="ofile", help="output (TSV)")
  parser.add_argument("--ids", help="code ID list (comma-separated)(e.g. 250.33)")
  parser.add_argument("--codesys", choices=medline.connect.Utils.CODESYSTEMS.keys(), default="RXNORM", help="code system")
  parser.add_argument("--api_host", default=medline.connect.Utils.API_HOST)
  parser.add_argument("--api_base_path", default=medline.connect.Utils.API_BASE_PATH)
  parser.add_argument("-v", "--verbose", action="count", default=0)
  args = parser.parse_args()

  logging.basicConfig(format='%(levelname)s:%(message)s', level=(logging.DEBUG if args.verbose>1 else logging.INFO))

  api_base_url='https://'+args.api_host+args.api_base_path

  fout = open(args.ofile, "w+") if args.ofile else sys.stdout

  t0=time.time()

  ids=[]
  if args.ifile:
    fin = open(args.ifile)
    while True:
      line = fin.readline()
      if not line: break
      ids.append(line.rstrip())
    fin.close()
  elif args.ids:
    ids = re.split('[, ]+', args.ids.strip())
  if len(ids)>0: logging.info(f"Input IDs: {len(ids)}")

  if args.op=="get_code":
    medline.connect.Utils.GetCode(ids, args.codesys.upper(), api_base_url, fout)

  else:
    parser.error(f"Operation invalid: {args.op}")

  logging.info(f"Elapsed time: {time.strftime('%Hh:%Mm:%Ss',time.gmtime(time.time()-t0))}")




#!/usr/bin/env python
"""
Monarch REST client 

https://monarchinitiative.org/page/services
https://monarchinitiative.org/docs/files/api-js.html
https://github.com/monarch-initiative/owlsim-v3
https://github.com/monarch-initiative/monarch-analysis

/search	searches over ontology terms via OntoQuest
/autocomplete	proxy for vocbaulary services autocomplete
/disease	disease info or page
/phenotype	phenotype info or page
/simsearch	OwlSim search using search profile of ontology classes

https://monarchinitiative.org/disease/OMIM_127750.json
https://monarchinitiative.org/phenotype/HP_0000003.json
https://monarchinitiative.org/compare/OMIM:270400/NCBIGene:5156,OMIM:249000,OMIM:194050.json

We use OwlSim for semantic matching when comparing two entities (such as genes,
diseases, or genotypes) with sets of attributes (phenotypes). URLs are of the form
/compare/:x/:y.json
where x can be either an entity or a list of phenotypes and y can be a list of
entities and/or sets of phenotypes.
This wraps a call to getGroupwiseSimilarity(x,y) in OwlSim. This works such that
given a query id (such as a gene, genotype, disease), and one or more target
identifiers, it will map each to it's respective phenotypes, and perform an OwlSim
comparison of the query to each target. You are permitted to mix query and target
types. For example, your query can be a disease, and the target(s) be a list of
gene(s), disease(s), phenotype(s), and/or genotype(s). You can indicate to union the
phenotypes of either the query or the target with a plus "+". Only one entity may be
supplied for the query, whereas multiple target entities are allowed (delimited by a
comma).  For details on owlsim, see http://owlsim.org

Paths:
/compare/:id1/:id2
/compare/:id1/:id2,id3,...idN
/compare/:id1+:id2/:id3,:id4,...idN
/compare/:id1/:id2+:id3,:id4,:id5+:id6,...,:idN

IC = Information Content
LCS = Least Common Subsumers
See OwlSim docs and pubs.
https://www.ncbi.nlm.nih.gov/pmc/articles/PMC3653101/
"""
###
import sys,os,re,json,argparse,time,logging
#
from .. import monarch
#
##############################################################################
if __name__=='__main__':
  epilog='''\
compare: idAs vs idBs, OwlSim phenotype-based similarity
Entity: gene|disease|genotype
Examples:
OMIM:612528 (human disease),
HP:0005978 (human disease: T2DM),
DOID:1612 (human disease: breast cancer),
NCBIGene:7084 (human gene),
NCBIGene:5156 (human gene),
NCBIGene:3257 (human gene),
NCBIGene:26564 (mouse gene)
'''
  parser = argparse.ArgumentParser(description="Monarch Initiative REST client", epilog=epilog)
  ops = [
        "get_disease",
        "get_disease_relationships",
        "get_phenotype",
        "get_gene",
        "compare_phenotypes"
	]
  parser.add_argument("op", choices=ops, help="operation")
  parser.add_argument("--i", dest="ifile", help="input ID file")
  parser.add_argument("--idAs", help="input IDs (comma-separated)")
  parser.add_argument("--idBs", help="input IDs (comma-separated)")
  parser.add_argument("--o", dest="ofile", help="output (TSV)")
  parser.add_argument("--api_host", default=monarch.API_HOST)
  parser.add_argument("--api_base_path", default=monarch.API_BASE_PATH)
  parser.add_argument("-v", "--verbose", dest="verbose", action="count", default=0)
  args = parser.parse_args()

  logging.basicConfig(format='%(levelname)s:%(message)s', level=(logging.DEBUG if args.verbose>1 else logging.INFO))

  base_url = 'https://'+args.api_host+args.api_base_path

  fout = open(args.ofile,"w") if args.ofile else sys.stdout

  t0=time.time()

  idAs=[]; idBs=[];
  if args.ifile:
    fin = open(args.ifile)
    while True:
      line = fin.readline()
      if not line: break
      if line.rstrip(): idAs.append(line.rstrip())
    fin.close()
  elif args.idAs:
    idAs = re.split(r'[,\s]+', args.idAs)
  logging.info("Input idAs: {len(idAs)}")
  if args.idBs:
    idBs = re.split(r'[,\s]+', args.idBs)
  logging.info("Input idBs: {len(idBs)}")

  if args.op=="get_disease":
    monarch.GetDisease(idAs, base_url, fout)

  elif args.op=="get_disease_relationships":
    monarch.GetDiseaseRelationships(idAs, base_url, fout)

  elif args.op=="get_phenotype":
    parser.error(f"Unimplemented operation: {args.op}")
    #monarch.GetPhenotype(idAs, base_url, fout)

  elif args.op=="get_gene":
    monarch.GetGene(idAs, base_url, fout)

  elif args.op=="compare_phenotypes":
    monarch.ComparePhenotypes(idAs, idBs, base_url, fout)

  else:
    parser.error(f"Invalid operation: {args.op}")

  logging.info(f"Elapsed time: {time.strftime('%Hh:%Mm:%Ss',time.gmtime(time.time()-t0))}")



#!/usr/bin/env python3
"""
Utilities for the AMP T2D REST API.
http://www.type2diabetesgenetics.org/
http://www.kp4cd.org/apis/t2d
http://52.54.103.84/kpn-kb-openapi/

DEPICT software (Pers, TH, et al., 2015)
"""
###
import sys,os,re,json,time,logging
#
from ...util import rest
#
#############################################################################
def ListTissues(base_url, fout):
  rval = rest.Utils.GetURL(base_url+'/graph/tissue/list/object', parse_json=True)
  tissues = rval["data"] if "data" in rval else []
  tags = None; n_out=0;
  for tissue in tissues:
    logging.debug(json.dumps(tissue, indent=2))
    if not tags:
      tags = tissue.keys()
      fout.write('\t'.join(tags)+'\n')
    vals = [str(tissue[tag]) if tag in tissue else '' for tag in tags]
    fout.write('\t'.join(vals)+'\n')
    n_out += 1
  logging.info("n_out: %d"%(n_out))

#############################################################################
def ListPhenotypes(base_url, fout):
  rval=rest.Utils.GetURL(base_url+'/graph/phenotype/list/object', parse_json=True)
  phenotypes = rval["data"] if "data" in rval else []
  tags = None; n_out=0;
  for phenotype in phenotypes:
    logging.debug(json.dumps(phenotype, indent=2))
    if not tags:
      tags = phenotype.keys()
      fout.write('\t'.join(tags)+'\n')
    vals = [str(phenotype[tag]) if tag in phenotype else '' for tag in tags]
    fout.write('\t'.join(vals)+'\n')
    n_out += 1
  logging.info("n_out: %d"%(n_out))

##############################################################################
def DepictGenePathway(base_url, gene, phenotype, max_pval, fout):
  url = base_url+('/testcalls/depict/genepathway/object?gene=%s&phenotype=%s&lt_value=%f'%(gene, phenotype, max_pval))
  rval = rest.Utils.GetURL(url, parse_json=True)
  pathways = rval["data"] if "data" in rval else []
  tags = None; n_out=0;
  for pathway in pathways:
    logging.debug(json.dumps(pathway, indent=2))
    if not tags:
      tags = pathway.keys()
      fout.write('\t'.join(tags)+'\n')
    vals = [str(pathway[tag]) if tag in pathway else '' for tag in tags]
    fout.write('\t'.join(vals)+'\n')
    n_out += 1
  logging.info("n_out: %d"%(n_out))

##############################################################################



#!/usr/bin/env python3
"""
Pandas/Tabulate Csv2Markdown.
Tabulate formatting https://pypi.org/project/tabulate/
"""
###
import sys,os,argparse,re,logging
import tabulate
import pandas as pd

# default=lambda(x: f"{x:.3f}"),
#############################################################################
if __name__=='__main__':
  FORMATS = ["plain", "simple", "github", "grid", "fancy_grid", "pipe", "orgtbl", "jira", "presto", "pretty", "psql", "rst", "mediawiki", "moinmoin", "youtrack", "html", "unsafehtml", "latex", "latex_raw", "latex_booktabs", "latex_longtable", "textile", "tsv"]
  parser = argparse.ArgumentParser(description='Pandas/Tabulate Csv2Markdown.')
  parser.add_argument("--i", dest="ifile", required=True, help="input (CSV|TSV)")
  parser.add_argument("--o", dest="ofile", help="output (HTML)")
  parser.add_argument("--csv", action="store_true", help="delimiter is comma")
  parser.add_argument("--tsv", action="store_true", help="delimiter is tab")
  parser.add_argument("--title", help="Markdown heading")
  parser.add_argument("--columns", help="Subset of columns to write (comma delimited)")
  parser.add_argument("--nrows", type=int)
  parser.add_argument("--skiprows", type=int)
  parser.add_argument("--numalign", choices=["center","right","left","decimal"], default="center")
  parser.add_argument("--stralign", choices=["center","right","left"], default="left")
  parser.add_argument("--format", choices=FORMATS, default="github", help="tabulate format (tablefmt)")
  parser.add_argument("--na_rep", default="", help="String representation of NaN")
  parser.add_argument("--float_format", help="Function(float) -> string.")
  parser.add_argument("-v", "--verbose", action="count", default=0)
  args = parser.parse_args()

  logging.basicConfig(format='%(levelname)s:%(message)s', level=(logging.DEBUG if args.verbose>1 else logging.INFO))

  fout = open(args.ofile, "w") if args.ofile else sys.stdout

  if args.csv: delim=','
  elif args.tsv: delim='\t'
  elif re.search('\.csv', args.ifile, re.I): delim=','
  else: delim='\t'

  title = args.title if args.title else f"Csv2Markdown: {os.path.basename(args.ifile)}"
  columns = re.split(r',', args.columns) if args.columns else None

  df = pd.read_csv(args.ifile, sep=delim, nrows=args.nrows, skiprows=args.skiprows)

  if columns is not None:
    df = df[columns]

  if args.na_rep is not None:
    df = df.fillna(args.na_rep)

  table_md = df.to_markdown(tablefmt=args.format)

  md = f"""
# {title}

{table_md}
"""

  fout.write(md)
  fout.close()




#!/usr/bin/env python3
"""
CFChemDb PostgreSql db client.
"""
import os,sys,argparse,re,time,logging

from ...cfde import cfchemdb
from ...util import yaml as util_yaml
from ...util import db as util_db

#############################################################################
if __name__=='__main__':
  parser = argparse.ArgumentParser(description="CFChemDb PostgreSql client utility", epilog="")
  ops = [
	"list_tables",
	"list_columns",
	"list_tables_rowCounts",
	"version",
	"get_structure",
	"list_structures",
	"list_structures2smiles",
	"meta_listdbs"
	]
  parser.add_argument("op", choices=ops, help="OPERATION (select one)")
  parser.add_argument("--i", dest="ifile", help="input ID file")
  parser.add_argument("--ids", help="input IDs (comma-separated)")
  parser.add_argument("--xref_type", help="xref ID type")
  parser.add_argument("--o", dest="ofile", help="output (TSV)")
  parser.add_argument("--dbhost")
  parser.add_argument("--dbport")
  parser.add_argument("--dbname")
  parser.add_argument("--dbusr")
  parser.add_argument("--dbpw")
  parser.add_argument("--param_file", default=os.environ['HOME']+"/.cfchemdb.yaml")
  parser.add_argument("--dbschema", default="public")
  parser.add_argument("-v", "--verbose", dest="verbose", action="count", default=0)
  parser.add_argument("-q", "--quiet", action="store_true", help="Suppress progress notification.")
  args = parser.parse_args()

  # logging.PROGRESS = 15 (custom)
  logging.basicConfig(format='%(levelname)s:%(message)s', level=(logging.DEBUG if args.verbose>0 else logging.ERROR if args.quiet else 15))

  if os.path.isfile(args.param_file):
    params = util_yaml.ReadParamFile(args.param_file)
  if args.dbhost: params['DBHOST'] = args.dbhost 
  if args.dbport: params['DBPORT'] = args.dbport 
  if args.dbname: params['DBNAME'] = args.dbname 
  if args.dbusr: params['DBUSR'] = args.dbusr 
  if args.dbpw: params['DBPW'] = args.dbpw 

  fout = open(args.ofile, "w+") if args.ofile else sys.stdout

  t0 = time.time()

  ids=[]
  if args.ifile:
    fin = open(args.ifile)
    while True:
      line = fin.readline()
      if not line: break
      ids.append(line.rstrip())
    logging.info('Input IDs: %d'%(len(ids)))
    fin.close()
  elif args.ids:
    ids = re.split(r'[,\s]+', args.ids)

  try:
    #dbcon = cfchemdb.Connect(params['DBHOST'], params['DBPORT'], params['DBNAME'], params['DBUSR'], params['DBPW'])
    dbcon = util_db.PostgreSqlConnect(dbhost=params['DBHOST'], dbport=params['DBPORT'], dbusr=params['DBUSR'], dbpw=params['DBPW'], dbname=params['DBNAME'])
  except Exception as e:
    logging.error(f"Connect failed: {e}")
    sys.exit(1)

  if args.op=='list_tables':
    cfchemdb.ListTables(dbcon, args.dbschema, fout)

  elif args.op=='list_tables_rowCounts':
    cfchemdb.ListTablesRowCounts(dbcon, args.dbschema, fout)

  elif args.op=='list_columns':
    cfchemdb.ListColumns(dbcon, args.dbschema, fout)

  elif args.op=='version':
    cfchemdb.Version(dbcon, args.dbschema, fout)

  elif args.op=='list_structures':
    cfchemdb.ListStructures(dbcon, args.dbschema, fout)

  elif args.op=='list_structures2smiles':
    cfchemdb.ListStructures2Smiles(dbcon, args.dbschema, fout)

  elif args.op=='get_structure':
    cfchemdb.GetStructure(dbcon, ids, fout)

  elif args.op=='meta_listdbs':
    cfchemdb.MetaListdbs(dbcon, fout)

  else:
    parser.error(f"Invalid operation: {args.op}")

  dbcon.close()
  logging.info(f"Elapsed time: {time.strftime('%Hh:%Mm:%Ss', time.gmtime(time.time()-t0))}")

